#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

from math import *
import string
import re
import os.path
import datetime
import traceback
import time

import qubtree
from Buttons import AddTaskButton
from Dialogs import *
import QUB
import CurveFitting
import QUBFuzzyFit

def EvensOdds(lst):
    evens = [x for i, x in enumerate(lst) if not (i%2)]
    odds  = [x for i, x in enumerate(lst) if (i%2)]
    return evens, odds

class Anon:
    def __init__(self, **kw):
        for field in kw.keys():
            self.__dict__[field] = kw[field]
    def __repr__(self):
        return str(self.__dict__)
    def __str__(self):
        return str(self.__dict__)

def children(tree, name=None):
    if name == None:
        child = tree.child
    else:
        child = tree.find(name)
    while not child.isNull:
        yield child
        if name == None:
            child = child.sibling
        else:
            child = child.nextSameName()


def DistHist(data, nbin=32, title="", xlabel="", ylabel=""):
    lo = min(data)
    hi = max(data)
    binw = (hi - lo) / nbin
    bin0m = lo + binw/2
    bins = [i*binw + bin0m for i in xrange(nbin)]
    bars = [0.0] * nbin
    for x in data:
        bin = min(nbin-1, int(floor((x-lo)/binw)))
        bars[bin] += 1.0
    hist = qubtree.Node('Histogram')
    hist['Title'].data = title
    hist['XLabel'].data = xlabel
    hist['YLabel'].data = ylabel
    hist['BinCount'].data = nbin
    hist['Bins'].data = bins
    hist['Bars'].data = bars
    return hist


def ChiSqr(data, fit, nfree):
    x2 = 0.0
    for i in range(len(data)):
        var = data[i] - fit[i]
        x2 += var * var
    return x2 / nfree

def Uncertainty(fitter, fitdata, deltapct):
    params = CurveFitting.componentToList(fitter.params[0])
    X = fitdata['X'].data
    data = fitdata['Y'].data
    ndata = len(data)
    nfree = ndata - len(params)
    chiSqrFit = ChiSqr(data, fitdata['Fit'].data, nfree)
    sigma = []
    for i, pval in enumerate(params):
        delta = abs(deltapct * pval)
        if delta == 0:
            delta = 0.1
        newparams = params[:]
        chiSqrOff = []
        for sign in [-1, 1]:
            newparams[i] = pval + sign * delta
            newfit = [fitter.func(newparams, xx) for xx in X]
            chiSqrOff.append( ChiSqr(data, newfit, nfree) )
        sigma.append( delta * sqrt(2.0 / (nfree * (chiSqrOff[0] + chiSqrOff[1] - 2.0*chiSqrFit))) )
    return sigma

class MultiFitter:
    def __init__(self):
        self.wiz = QUB.CreateWizard()
        self.wiz.caption = 'MultiFit'
        self.wiz.buttons = ['Grad', 'Fuzzy', 'Params', 'Fit Avg', 'Fit 1st', 'Fit Each', 'Done']
        self.wizActions = {'Grad'          : self.SetupGrad,
                           'Fuzzy'         : self.SetupFuzzy,
                           'DumbG'         : self.SetupDumbGrad,
                           'Params'        : self.EditParams,
                           'Fit Avg'       : self.FitAvg,
                           'Fit 1st'       : self.FitFirst,
                           'Fit Each'      : self.FitEach}
        self.properties = QUB.Profiles['MFit'].active
        def initProp(name, val, base=self.properties):
            if base.find(name).isNull:
                base[name].data = val
        initProp('maxiter', 30)
        initProp('precision', 0.01)
        initProp('algo', CurveFitting.MARQUARDT)
        initProp('mag', 0.05)
        initProp('nfuzz', 10)
        initProp('maxnfuzz', 20)
        initProp('rad', 0.5)
        initProp('decay', 0.95)
        initProp('ncircle', 64)
        initProp('linmean', 0.5)
        initProp('lincount', 64)
        initProp('decimation', 1)
        initProp('eqn', 'Y0 + M*x + A*(exp(x/T) - 1.0)')
        initProp('Y0', 0.0, self.properties['Params'])
        initProp('M', -0.3, self.properties['Params'])
        initProp('A', .5, self.properties['Params'])
        initProp('T', 10, self.properties['Params'])
        self.fitter = None
        self.Display()
        self.InitData()
    def Display(self):
        if not self.fitter:
            msg = 'Please choose a method\r\n(Grad, Fuzzy, or DumbG)'
        else:
            msg = str(self.properties['eqn'].data)
            msg += '\r\nss residual of avg: %f' % self.avgfitdata['SSD'].data[0]
            msg += '\r\nrms residual: %f' % self.avgfitdata['rms'].data[0]
            for param in children(self.properties['Params']):
                msg += '\r\n\t%s\t= %s' % (param.name, param.data)
        self.wiz.message = msg
    def ReadButtons(self):
        self.wiz.show()
        btn = None
        while btn != 'Done':
            btn = self.wiz.waitForButton()
            if btn != 'Done':
                try:
                    self.wizActions[btn]()
                except:
                    traceback.print_exc()
        self.wiz.hide()
    def InitData(self):
        if QUB.GetDataSource() == 'file list':
            QUB.SetDataSource('file')
        dss = QUB.getChosenData()
        self.dataset = dss['DataSet']
        self.datadesc = dss['DataDescription']
        self.dataFile = QUB.DataView.currData
        segs, n = [], 100000000
        for seg in children(self.dataset, 'Segment'):
            segs.append( seg )
            count = seg.data[1] - seg.data[0] + 1
            if count < n:
                n = count
        sampling = self.dataset['sampling'].data[0] * 1.0e-3
        xx = qubtree.Node('X')
        xx.data = [i*sampling for i in xrange(n)]
        self.avgfitdata = qubtree.Node('FitData')
        self.avgfitdata.appendClone( xx )
        avgy = [0.0] * n
        self.selfitdata = []
        for seg in segs:
            fitdata = qubtree.Node('FitData')
            fitdata.appendClone( xx )
            QUB.fillData( seg )
            fitdata['Y'].data = seg['Channel'].data
            for i, y in enumerate(seg['Channel'].data):
                if i < n:
                    avgy[i] += y
            fitdata['Fit'].data = [0.0] * n
            fitdata['SSD'].data = 0.0
            self.selfitdata.append( fitdata )
            seg.remove(seg['Channel'])
        avgscale = 1.0 / len(segs)
        for i in xrange(n):
            avgy[i] *= avgscale
        self.avgfitdata['Y'].data = avgy
        self.avgfitdata['Fit'].data = [0.0] * n
        self.avgfitdata['SSD'].data = 0.0
    def UpdateCurve(self, fitdata):
        ssd = 0.0
        X, Y = fitdata['X'].data, fitdata['Y'].data
        curve = fitdata['Fit'].data
        fitter = self.fitter
        paramvals = CurveFitting.componentToList(fitter.params[0])
        for i in range(len(curve)):
            curve[i] = fitter.func(paramvals, X[i])
            diff = curve[i] - Y[i]
            ssd += diff * diff
        fitdata['SSD'].data = ssd
        fitdata['rms'].data = sqrt(ssd / len(curve))
    def DisplayCurve(self, fitdata, sel):
        seg = qubtree.Node('Segment')
        seg.data = sel
        seg.appendClone(fitdata['Fit']).name = 'Channel'
        self.dataFile.setFitCurve(seg)
    def SetupGrad(self):
        fitter = CurveFitting.CustomFitter()
        fitter.algo = self.properties['algo'].data[0]
        fitter.setCustomEqn(str(self.properties['eqn'].data))
        fitter.promptOptions()
        self.properties['algo'].data = fitter.algo
        self.SetupFitter(fitter)
    def SetupFuzzy(self):
        fitter = QUBFuzzyFit.FuzzyFitter()
        fitter.setEqn(str(self.properties['eqn'].data))
        options = ['mag', 'nfuzz', 'maxnfuzz', 'decimation']
        for x in options:
            fitter.__dict__[x] = self.properties[x].data[0]
        fitter.promptOptions()
        for x in options:
            self.properties[x].data = fitter.__dict__[x]
        self.SetupFitter(fitter)
    def SetupDumbGrad(self):
        fitter = QUBFuzzyFit.DumbGradFitter()
        fitter.setEqn(str(self.properties['eqn'].data))
        options = ['rad', 'decay', 'linmean', 'ncircle', 'lincount', 'decimation']
        for x in options:
            fitter.__dict__[x] = self.properties[x].data[0]
        fitter.promptOptions()
        for x in options:
            self.properties[x].data = fitter.__dict__[x]
        self.SetupFitter(fitter)
    def SetupFitter(self, fitter):
        self.properties['eqn'].data = fitter.eqn
        self.fitter = fitter
        self.SetupParams()
        self.UpdateCurve(self.avgfitdata)
        self.DisplayCurve(self.avgfitdata, self.dataset['Segment'].data)
        self.Display()
    def SetupParams(self):
        for param in children(self.fitter.params[0]):
            mypar = self.properties['Params'][param.name]
            if len(mypar.data) == 0:
                mypar.data = param.data[0]
            else:
                param.data[0] = mypar.data[0]
    def EditParams(self):
        names = [param.name for param in children(self.fitter.params[0])]
        params = self.properties['Params']
        options = [DialogText(name, name, params[name].data[0], acceptFloat) for name in names]
        response = QUB.Dialog('MultiFit: Edit Params', ['OK', 'Cancel'], options)
        if response['button'] == 'OK':
            for param in children(params):
                try:
                    param.data = response[param.name]
                except:
                    pass
            self.SetupParams()
            self.UpdateCurve(self.avgfitdata)
            self.DisplayCurve(self.avgfitdata, self.dataset['Segment'].data)
            self.Display()
    def FitAvg(self):
        if self.PromptFit():
            self.wiz.message = 'Fitting...'
            self.FitOne(self.avgfitdata, self.properties['maxiter'].data[0], self.properties['precision'].data[0])
            for param in children(self.fitter.params[0]):
                self.properties['Params'][param.name].data = param.data[0]
            self.DisplayCurve(self.avgfitdata, self.dataset['Segment'].data)
            self.Display()
    def FitFirst(self):
        if self.PromptFit():
            self.wiz.message = 'Fitting...'
            self.FitOne(self.selfitdata[0], self.properties['maxiter'].data[0], self.properties['precision'].data[0])
            for param in children(self.fitter.params[0]):
                self.properties['Params'][param.name].data = param.data[0]
            self.DisplayCurve(self.selfitdata[0], self.dataset['Segment'].data)
            self.Display()
    def FitEach(self):
        if not self.PromptFit():
            return
        self.SetupParams()
        paramnames = [param.name for param in children(self.fitter.params[0])]
        maxiter = self.properties['maxiter'].data[0]
        precision = self.properties['precision'].data[0]
        result = qubtree.Node('MFit Result')
        result['ResultType'].data = 'MFit'
        result.appendClone(self.datadesc)
        ds = result.appendClone(self.dataset)
        for i, seg, fitdata in zip(xrange(len(self.selfitdata)), children(ds, 'Segment'), self.selfitdata):
            self.wiz.message = 'Fitting selection %d...' % (i+1)
            # self.SetupParams() # don't reload initial param values; use finals from last selection
            self.FitOne(fitdata, maxiter, precision)
            seg['residual'].data = fitdata['SSD'].data[0]
            seg['rms residual'].data = fitdata['rms'].data[0]
            for name in paramnames:
                seg[name].data = self.fitter.params[0][name].data[0]
            seg['HistogramGroup'].append( DistHist([aa - bb for aa, bb in zip(fitdata['Y'].data, fitdata['Fit'].data)],
                                                   title="Selection %d"%(i+1), xlabel="Yi - Fi", ylabel="count") )
            sigma = Uncertainty(self.fitter, fitdata, self.properties['mag'].data[0])
            for i, name in enumerate(paramnames):
                seg['%s std' % name].data = sigma[i]
            self.DisplayCurve(fitdata, seg.data)
        self.dataFile.addResults(result)
        self.Display()
    def PromptFit(self):
        params = self.properties
        options = [DialogText('maxiter', 'Max iterations:', params['maxiter'].data[0], acceptIntGreaterThan(0)),
                   DialogText('precision', 'Precision:', params['precision'].data[0], acceptFloatGreaterThan(0))]
        response = QUB.Dialog('MultiFit: Fit', ['OK', 'Cancel'], options)
        params['maxiter'].data = response['maxiter']
        params['precision'].data = response['precision']
        return (response['button'] == 'OK')
    def FitOne(self, fitdata, maxiter, precision):
        self.fitter.useData(fitdata)
        self.fitter.fit(maxiter, precision)
        self.UpdateCurve(fitdata)

def MultiFit(props):
    try:
        MultiFitter().ReadButtons()
    except:
        traceback.print_exc()


#AddTaskButton( 'MultiFit', # button text
#               MultiFit,   # function
#               'MFit',       # short text for small buttons
#               0,            # show settings dialog on left-click?
#               0,            # show settings dialog on right-click?
#               "Curve fitting for multiple data selections",   # tool-tip
#               "Curve fitting for multiple data selections",   # description in "Edit Buttons"
#               "Properties:", # caption for the settings dialog
#               [],
#               'Modeling')



